import descriptors from '../internals/descriptors.js';
import toIndexedObject from '../internals/to-indexed-object.js';
import objectGetOwnPropertyDescriptor from '../internals/object-get-own-property-descriptor.js';
import ownKeys from '../internals/own-keys.js';
import _export from '../internals/export.js';
import createProperty from '../internals/create-property.js';

// `Object.getOwnPropertyDescriptors` method
// https://tc39.github.io/ecma262/#sec-object.getownpropertydescriptors
_export({ target: 'Object', stat: true, sham: !descriptors }, {
  getOwnPropertyDescriptors: function getOwnPropertyDescriptors(object) {
    var O = toIndexedObject(object);
    var getOwnPropertyDescriptor = objectGetOwnPropertyDescriptor.f;
    var keys = ownKeys(O);
    var result = {};
    var index = 0;
    var key, descriptor;
    while (keys.length > index) {
      descriptor = getOwnPropertyDescriptor(O, key = keys[index++]);
      if (descriptor !== undefined) createProperty(result, key, descriptor);
    }
    return result;
  }
});
