import { getGlobalClassNames, HighContrastSelector, HighContrastSelectorBlack, HighContrastSelectorWhite } from '@uifabric/styling';
import defineProperty from '../../../node_modules/@babel/runtime/helpers/defineProperty.js';
import '../../../node_modules/core-js/modules/es.string.link.js';

var GlobalClassNames = {
  root: 'ms-Link'
};
var getStyles = function getStyles(props) {
  var _selectors3;

  var className = props.className,
      isButton = props.isButton,
      isDisabled = props.isDisabled,
      theme = props.theme;
  var semanticColors = theme.semanticColors; // Tokens

  var linkColor = semanticColors.link;
  var linkInteractedColor = semanticColors.linkHovered;
  var linkDisabledColor = semanticColors.disabledText;
  var focusBorderColor = semanticColors.focusBorder;
  var classNames = getGlobalClassNames(GlobalClassNames, theme);
  return {
    root: [classNames.root, theme.fonts.medium, {
      color: linkColor,
      outline: 'none',
      fontSize: 'inherit',
      fontWeight: 'inherit',
      textDecoration: 'none',
      selectors: defineProperty({
        '.ms-Fabric--isFocusVisible &:focus': {
          // Can't use getFocusStyle because it doesn't support wrapping links
          // https://github.com/OfficeDev/office-ui-fabric-react/issues/4883#issuecomment-406743543
          // Using box-shadow and outline allows the focus rect to wrap links that span multiple lines
          // and helps the focus rect avoid getting clipped.
          boxShadow: "0 0 0 1px ".concat(focusBorderColor, " inset"),
          outline: "1px auto ".concat(focusBorderColor),
          selectors: defineProperty({}, HighContrastSelector, {
            outline: '1px solid WindowText'
          })
        }
      }, HighContrastSelector, {
        // For IE high contrast mode
        borderBottom: 'none'
      })
    }, isButton && {
      background: 'none',
      backgroundColor: 'transparent',
      border: 'none',
      cursor: 'pointer',
      display: 'inline',
      margin: 0,
      overflow: 'inherit',
      padding: 0,
      textAlign: 'left',
      textOverflow: 'inherit',
      userSelect: 'text',
      borderBottom: '1px solid transparent',
      selectors: (_selectors3 = {}, defineProperty(_selectors3, HighContrastSelectorBlack, {
        color: '#FFFF00'
      }), defineProperty(_selectors3, HighContrastSelectorWhite, {
        color: '#00009F'
      }), _selectors3)
    }, isDisabled && ['is-disabled', {
      color: linkDisabledColor,
      cursor: 'default'
    }, {
      selectors: {
        '&:link, &:visited': {
          pointerEvents: 'none'
        }
      }
    }], !isDisabled && {
      selectors: {
        '&:active, &:hover, &:active:hover': {
          color: linkInteractedColor,
          textDecoration: 'underline'
        },
        '&:focus': {
          color: linkColor
        }
      }
    }, classNames.root, className]
  };
};

export { getStyles };
