import '../../../node_modules/core-js/modules/es.object.to-string.js';
import '../../../node_modules/core-js/modules/es.regexp.to-string.js';
import classCallCheck from '../../../node_modules/@babel/runtime/helpers/classCallCheck.js';
import createClass from '../../../node_modules/@babel/runtime/helpers/createClass.js';
import { EventGroup, getId } from '@uifabric-vue/utilities';
import '../../../node_modules/core-js/modules/es.symbol.js';
import '../../../node_modules/core-js/modules/es.array.filter.js';
import '../../../node_modules/core-js/modules/es.array.for-each.js';
import '../../../node_modules/core-js/modules/es.object.get-own-property-descriptor.js';
import '../../../node_modules/core-js/modules/es.object.get-own-property-descriptors.js';
import '../../../node_modules/core-js/modules/es.object.keys.js';
import '../../../node_modules/core-js/modules/web.dom-collections.for-each.js';
import defineProperty from '../../../node_modules/@babel/runtime/helpers/defineProperty.js';
import '../../../node_modules/core-js/modules/es.array.map.js';
import toConsumableArray from '../../../node_modules/@babel/runtime/helpers/toConsumableArray.js';
import { KeytipEvents } from './KeytipConstants.js';

function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys(Object(source), true).forEach(function (key) { defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }
/**
 * This class is responsible for handling registering, updating, and unregistering of keytips
 */

var KeytipManager = /*#__PURE__*/function () {
  function KeytipManager() {
    classCallCheck(this, KeytipManager);

    this.keytips = {};
    this.persistedKeytips = {};
    this.sequenceMapping = {}; // This is (and should be) updated and kept in sync
    // with the inKeytipMode in KeytipLayer.

    this.inKeytipMode = false; // Boolean that gets checked before entering keytip mode by the KeytipLayer
    // Used for an override in special cases (e.g. Disable entering keytip mode when a modal is shown)

    this.shouldEnterKeytipMode = true; // Boolean to indicate whether to delay firing an event to update subscribers of
    // keytip data changed.

    this.delayUpdatingKeytipChange = false;
  }
  /**
   * Static function to get singleton KeytipManager instance
   *
   * @returns {KeytipManager} Singleton KeytipManager instance
   */


  createClass(KeytipManager, [{
    key: "init",

    /**
     * Initialization code to set set parameters to define
     * how the KeytipManager handles keytip data.
     *
     * @param delayUpdatingKeytipChange - T/F if we should delay notifiying keytip subscribers
     * of keytip changes
     */
    value: function init(delayUpdatingKeytipChange) {
      this.delayUpdatingKeytipChange = delayUpdatingKeytipChange;
    }
    /**
     * Registers a keytip
     *
     * @param keytipProps - Keytip to register
     * @param persisted - T/F if this keytip should be persisted, default is false
     * @returns {string} Unique ID for this keytip
     */

  }, {
    key: "register",
    value: function register(keytipProps) {
      var persisted = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
      var props = keytipProps;

      if (!persisted) {
        // Add the overflowSetSequence if necessary
        props = this.addParentOverflow(keytipProps);
        this.sequenceMapping[props.keySequences.toString()] = props;
      } // Create a unique keytip


      var uniqueKeytip = this._getUniqueKtp(props); // Add to dictionary


      persisted ? this.persistedKeytips[uniqueKeytip.uniqueID] = uniqueKeytip : this.keytips[uniqueKeytip.uniqueID] = uniqueKeytip; // We only want to add something new if we are currently showing keytip mode

      if (this.inKeytipMode || !this.delayUpdatingKeytipChange) {
        var event = persisted ? KeytipEvents.PERSISTED_KEYTIP_ADDED : KeytipEvents.KEYTIP_ADDED;
        EventGroup.raise(this, event, {
          keytip: props,
          uniqueID: uniqueKeytip.uniqueID
        });
      }

      return uniqueKeytip.uniqueID;
    }
    /**
     * Update a keytip
     *
     * @param keytipProps - Keytip to update
     * @param uniqueID - Unique ID of this keytip
     */

  }, {
    key: "update",
    value: function update(keytipProps, uniqueID) {
      var newKeytipProps = this.addParentOverflow(keytipProps);

      var uniqueKeytip = this._getUniqueKtp(newKeytipProps, uniqueID);

      var oldKeyTip = this.keytips[uniqueID];

      if (oldKeyTip) {
        // Update everything except 'visible'
        uniqueKeytip.keytip.visible = oldKeyTip.keytip.visible; // Update keytip in this.keytips

        this.keytips[uniqueID] = uniqueKeytip; // Update the sequence to be up to date

        delete this.sequenceMapping[oldKeyTip.keytip.keySequences.toString()];
        this.sequenceMapping[uniqueKeytip.keytip.keySequences.toString()] = uniqueKeytip.keytip; // Raise event only if we are currently in keytip mode

        if (this.inKeytipMode || !this.delayUpdatingKeytipChange) {
          EventGroup.raise(this, KeytipEvents.KEYTIP_UPDATED, {
            keytip: uniqueKeytip.keytip,
            uniqueID: uniqueKeytip.uniqueID
          });
        }
      }
    }
    /**
     * Unregisters a keytip
     *
     * @param keytipToRemove - any of the keytip to remove
     * @param uniqueID - Unique ID of this keytip
     * @param persisted - T/F if this keytip should be persisted, default is false
     */

  }, {
    key: "unregister",
    value: function unregister(keytipToRemove, uniqueID) {
      var persisted = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;
      persisted ? delete this.persistedKeytips[uniqueID] : delete this.keytips[uniqueID];
      !persisted && delete this.sequenceMapping[keytipToRemove.keySequences.toString()];
      var event = persisted ? KeytipEvents.PERSISTED_KEYTIP_REMOVED : KeytipEvents.KEYTIP_REMOVED; // Update keytips only if we're in keytip mode

      if (this.inKeytipMode || !this.delayUpdatingKeytipChange) {
        EventGroup.raise(this, event, {
          keytip: keytipToRemove,
          uniqueID: uniqueID
        });
      }
    }
    /**
     * Manual call to enter keytip mode
     */

  }, {
    key: "enterKeytipMode",
    value: function enterKeytipMode() {
      EventGroup.raise(this, KeytipEvents.ENTER_KEYTIP_MODE);
    }
    /**
     * Manual call to exit keytip mode
     */

  }, {
    key: "exitKeytipMode",
    value: function exitKeytipMode() {
      EventGroup.raise(this, KeytipEvents.EXIT_KEYTIP_MODE);
    }
    /**
     * Gets all any from this.keytips
     *
     * @returns {any[]} All keytips stored in the manager
     */

  }, {
    key: "getKeytips",
    value: function getKeytips() {
      var _this = this;

      return Object.keys(this.keytips).map(function (key) {
        return _this.keytips[key].keytip;
      });
    }
    /**
     * Adds the overflowSetSequence to the keytipProps if its parent keytip also has it
     *
     * @param keytipProps - Keytip props to add overflowSetSequence to if necessary
     * @returns {any} - Modified keytip props, if needed to be modified
     */

  }, {
    key: "addParentOverflow",
    value: function addParentOverflow(keytipProps) {
      var fullSequence = toConsumableArray(keytipProps.keySequences);

      fullSequence.pop();

      if (fullSequence.length !== 0) {
        var parentKeytip = this.sequenceMapping[fullSequence.toString()];

        if (parentKeytip && parentKeytip.overflowSetSequence) {
          return _objectSpread(_objectSpread({}, keytipProps), {}, {
            overflowSetSequence: parentKeytip.overflowSetSequence
          });
        }
      }

      return keytipProps;
    }
    /**
     * Public function to bind for overflow items that have a submenu
     *
     * @param overflowButtonSequences
     * @param keytipSequences
     */

  }, {
    key: "menuExecute",
    value: function menuExecute(overflowButtonSequences, keytipSequences) {
      EventGroup.raise(this, KeytipEvents.PERSISTED_KEYTIP_EXECUTE, {
        overflowButtonSequences: overflowButtonSequences,
        keytipSequences: keytipSequences
      });
    }
    /**
     * Creates an IUniqueKeytip object
     *
     * @param keytipProps - any
     * @param uniqueID - Unique ID, will default to the next unique ID if not passed
     * @returns {IUniqueKeytip} IUniqueKeytip object
     */

  }, {
    key: "_getUniqueKtp",
    value: function _getUniqueKtp(keytipProps) {
      var uniqueID = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : getId();
      return {
        keytip: _objectSpread({}, keytipProps),
        uniqueID: uniqueID
      };
    }
  }], [{
    key: "getInstance",
    value: function getInstance() {
      return this._instance;
    }
  }]);

  return KeytipManager;
}();
KeytipManager._instance = new KeytipManager();

export { KeytipManager };
